import numpy as np
import os, sys, copy, itertools
import yaml
import itertools
# WISDEM 
from wisdem.aeroelasticse.CaseGen_General import CaseGen_General, save_case_matrix, save_case_matrix_yaml
from wisdem.aeroelasticse.pyIECWind import pyIECWind_extreme, pyIECWind_turb
from wisdem.aeroelasticse.Util import FileTools

# ROSCO 
from ROSCO_toolbox import controller as ROSCO_controller
from ROSCO_toolbox import turbine as ROSCO_turbine
from ROSCO_toolbox import utilities as ROSCO_Utilities

class CaseGen_Control():
    '''
    Python class to ease controller parameter studies. A lot of this is very similar to CaseGen_IEC, 
    but a little bit of control flavor is added!.

    Parameters:
    parameter_filename: str
        controller tuning input yaml file
    '''

    def __init__(self, parameter_filename):

        # Load controller parameter file
        inps = yaml.safe_load(open(parameter_filename))
        self.path_params = inps['path_params']
        self.turbine_params = inps['turbine_params']
        self.controller_params = inps['controller_params']

        # Turbulent Wind Defaults (NTM)
        self.seed = np.random.uniform(1, 1e8)
        self.Turbulence_Class = 'B'  # IEC Turbulence Class
        self.IEC_WindType = 'NTM'
        self.z_hub = 90.  # wind turbine hub height (m)
        self.D = 126.  # rotor diameter (m)
        self.PLExp = 0.2
        self.AnalysisTime = 720.
        self.debug_level = 0
        self.overwrite = True

        # Cases
        case_inputs={}

    def gen_turbwind(self, Uref):
        '''
        Generate turbulent wind files. 

        Parameters
        ----------
        Uref: list
            list of wind speeds 
        '''
        # Load pyIECWind_Turb
        turbwind = pyIECWind_turb()
        turbwind.seed                = self.seed
        turbwind.Turbulence_Class    = self.Turbulence_Class
        turbwind.z_hub               = self.z_hub
        turbwind.D                   = self.D
        turbwind.PLExp               = self.PLExp
        turbwind.AnalysisTime        = self.AnalysisTime
        turbwind.debug_level         = self.debug_level
        turbwind.overwrite           = self.overwrite
        # User defined
        turbwind.outdir              = self.wind_dir
        turbwind.case_name           = self.case_name_base
        turbwind.Turbsim_exe         = self.Turbsim_exe
        turbwind.debug_level         = self.debug_level

        for U in Uref:
            wind_file, wind_file_type  = turbwind.execute(self.IEC_WindType, U)

        return wind_file, wind_file_type



    def gen_control_cases(self, input_params, DISCON_params, values, group):
        '''
        Generate control case input dictionary for controller parameters

        Parameters
        ----------
        input_params: list
            List of parameters that exist in the controller tuning yaml file that will be modified
        DISCON_params: list
            List of parameters in DISCON.IN that will be changed by the paremeters in input_params
        values: list
            List with nested lists of floats that correspond to the input_params. 
                - i.e. [[0.1],[0.5]]
        group: int
            Group number for case_inputs

        Returns
        -------
        case_inputs: dict
            Dictionary of case inputs for WISDEM's CaseGen_General
        tuning_inputs: dict
            Dictionary containing yaml input tuning parameters
        '''
        # Load turbine
        turbine = ROSCO_turbine.Turbine(self.turbine_params)
        turbine.load_from_fast(self.path_params['FAST_InputFile'],
                               self.path_params['FAST_directory'], dev_branch=True)
        # if not os.path.exists(self.path_params['rotor_performance_filename']):
        #     R.write_rotor_performance(turbine,txt_filename=path_params['rotor_performance_filename'])

        # Check for flaps
        if 'zeta_flp' in input_params or 'omega_flp' in input_params:
            self.controller_params['Flp_Mode'] = 2
            turbine.load_blade_info()

        # Create tuning value matrix
        tuning_matrix = list(itertools.product(*values))

        # Generate cases 
        case_inputs = {}
        tuning_inputs = {key: [] for key in input_params}
        DISCON_vals_list = {key: [] for key in DISCON_params}
        for tuning_params in tuning_matrix:
            for inp_ind, inp_p in enumerate(input_params):
                # Modify controller parameters
                self.controller_params[inp_p] = tuning_params[inp_ind]
                # tuning_inputs[inp_p].append(tuning_params[inp_ind])

            # Initialize and tune controller
            controller = ROSCO_controller.Controller(self.controller_params)
            controller.tune_controller(turbine)

            # Get DISCON input parameters
            data_processing = ROSCO_Utilities.DataProcessing()
            DISCON_inputs = data_processing.DISCON_dict(turbine, controller)

            # Check for invalid controller inputs
            if (any(item in ['VS_KP', 'VS_KI'] for item in DISCON_params) and
                any(DISCON_inputs[DISCON_p][0] > 0 for DISCON_p in DISCON_params)):
                    print('WARNING: Control gains cannot be greater than zero and are for {} = {}'
                        .format([param for param in input_params],  tuning_params))
            elif (any(item in ['Flp_Kp', 'Flp_Ki'] for item in DISCON_params) and
                any(DISCON_inputs[DISCON_p][0] < 0 for DISCON_p in DISCON_params)):
                    print('WARNING: Control gains cannot be less than zero and are for {} = {}'
                        .format([param for param in input_params],  tuning_params))
            # Save DISCON and tuning inputs
            else:
                for DISCON_p in DISCON_params:
                    DISCON_vals = DISCON_inputs[DISCON_p]
                    DISCON_vals_list[DISCON_p].append(DISCON_vals)
                for inp_ind, inp_p in enumerate(input_params):
                    tuning_inputs[inp_p].append(tuning_params[inp_ind]) 
        # Write case_inputs
        for DISCON_p in DISCON_params:
            case_inputs[('DISCON_in', DISCON_p)] = {'vals': DISCON_vals_list[DISCON_p], 
                                                    'group': group}

        return case_inputs, tuning_inputs

def append_case_matrix_yaml(file_dir, file_name, append_dict, name_base, package=1):
    '''
    Append the case_matrix_yaml with some more dictionary entries
        - probably controller tuning inputs

    Parameters
    ----------
    file_dir: str
        yaml file directory
    file_name: str
        yaml file name
    append_dict: dict
        dictionary to append to yaml file
    name_base: str
        base name of parameter to be appended { X in yaml_data[(X,'param)] }
    '''
    yaml_data = FileTools.load_yaml(os.path.join(file_dir, file_name), package=package)

    for key in append_dict.keys():
        rep_list = []
        for val in append_dict[key]:
            if type(val) in [np.float32, np.float64, np.single, np.double, np.longdouble]:
                val = float(val)
            elif type(val) in [np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64, np.intc, np.uintc, np.uint]:
                val = int(val)
            elif type(val) in [np.array, np.ndarray]:
                val = val.tolist()
            elif type(val) in [np.str_]:
                val = str(val)
            rep_list.append(val)
    
        yaml_data[(name_base, key)] = rep_list

    FileTools.save_yaml(file_dir, file_name, yaml_data)



from __future__ import print_function
import os, sys, time, shutil
from functools import partial
import multiprocessing as mp
import numpy as np
import yaml
try:
    import ruamel_yaml as ry
except:
    try:
        import ruamel.yaml as ry
    except:
        raise ImportError('No module named ruamel.yaml or ruamel_yaml')

import matplotlib.pyplot as plt
import pandas as pd

from ROSCO_toolbox.utilities import FAST_IO

from pCrunch import Analysis, pdTools


class FAST_Processing(object):
    '''
    A class with tools to post process batch OpenFAST output data
    '''

    def __init__(self, **kwargs):
        # Optional population class attributes from key word arguments
        self.OpenFAST_outfile_list = [[]] # list of lists containing absolute path to fast output files. Each inner list corresponds to a dataset to compare, and should be of equal length
        self.dataset_names = []       # (Optional) N labels that identify each dataset
        # (Optional) AeroelasticSE case matrix text file. Used to generated descriptions of IEC DLCS
        self.fname_case_matrix = ''

        # Analysis Control
        self.parallel_analysis = False  # True/False; Perform post processing in parallel
        self.parallel_cores = None      # None/Int>0; (Only used if parallel_analysis==True) number of parallel cores, if None, multiprocessing will use the maximum available
        self.verbose = False            # True/False; Enable/Disable non-error message outputs to screen

        # Analysis Options
        self.t0 = None       # float>=0    ; start time to include in analysis
        self.tf = None  # float>=0,-1 ; end time to include in analysis

        # Load ranking
        self.ranking_vars = [   ["RotSpeed"], 
                                ["TipDxc1", "TipDxc2", "TipDxc3"], 
                                ["TipDyc1", "TipDyc2", "TipDyc3"], 
                                ['RootMyb1', 'RootMyb2', 'RootMyb3'], 
                                ['RootMxb1', 'RootMxb2', 'RootMxb3'],
                                ['TwrBsFyt'],
                                ]  # List of lists
        self.ranking_stats = [  'max',
                                'max',
                                'max',
                                'max',
                                'max',
                                    ] # should be same length as ranking_vars
        
        self.DEL_info = None  # [('RootMyb1', 10), ('RootMyb2', 10), ('RootMyb3', 10)]

        # Save settings
        self.results_dir       = 'temp_results'
        self.save_LoadRanking  = False  # NJA - does not exist yet
        self.save_SummaryStats = False

        # Load kwargs
        for k, w in kwargs.items():
            try:
                setattr(self, k, w)
            except:
                pass

        # Setup multiprocessing:
        if self.parallel_analysis:
            # Make sure multi-processing cores are valid
            if not self.parallel_cores:
                self.parallel_cores = mp.cpu_count()
            elif self.parallel_cores == 1:
                self.parallel_analysis = False

        # Check for save directory
        if self.save_LoadRanking or self.save_SummaryStats:
            if not os.path.exists(self.results_dir):
                os.makedirs(self.results_dir)


        super(FAST_Processing, self).__init__()

        '''
        # Analysis Control
        self.plot_TimeSeries = False
        self.plot_FFT = False
        # self.plot_LoadRanking      = False
        
        # [str]       ; (Only used if plot_TimeSeries or plot_FFT = True) list of OpenFAST channels to plot
        self.plot_vars = []
        # [str]       ; (Optional)(Only used if plot_TimeSeries or plot_FFT = True) list axis labels for OpenFAST channels
        self.plot_vars_labels = []
        # float>0.    ; (Only used if plot_FFT = True) maximum frequency on x axis for FFT plots
        self.fft_x_max = 1.
        # True/False  ; (Only used if plot_FFT = True) Include 1P, 3P, 6P bands on FFT plots
        self.fft_show_RtSpeed = True
        # [floats]    ; (Optional)(Only used if plot_FFT = True) Additional frequencies to label on the FFT
        self.fft_include_f = []
        # [str]       ; (Optional)(Only used if plot_FFT = True) Legend names for additional frequencies to label on the FFT
        self.fft_include_f_labels = []
        # [str]       ; (Only used if save_LoadRanking = True) list of OpenFAST channels to perform load ranking on
        self.ranking_vars = []
        # str         ; (Only used if save_LoadRanking = True) statistic to rank loads by (min, max, mean, std, maxabs)
        self.ranking_statistic = 'maxabs'
        # int         ; (Only used if save_LoadRanking = True) number of loads to list per channel
        self.ranking_n_cases = 5
        '''

    def batch_processing(self):
        '''
        Run a full batch processing case!
        '''
        # ------------------ Input consistancy checks ------------------ #
        # Do we have a list of data?
        N = len(self.OpenFAST_outfile_list)
        if N == 0:
            raise ValueError('Output files not defined! Populate: "FastPost.OpenFAST_outfile_list". \n Quitting FAST_Processing.')
            

        # Do all the files exist?
        files_exist = True
        for i, flist in enumerate(self.OpenFAST_outfile_list):
            if isinstance(flist, str):
                if not os.path.exists(flist):
                    print('Warning! File "{}" does not exist.'.format(
                        flist))
                    self.OpenFAST_outfile_list.remove(flist)
            elif isinstance(flist, list):
                for fname in flist:
                    if not os.path.exists(fname):
                        files_exist = False
                        if len(self.dataset_names) > 0:
                            print('Warning! File "{}" from {} does not exist.'.format(
                                fname, self.dataset_names[i]))
                            flist.remove(fname)
                        else:
                            print('Warning! File "{}" from dataset {} of {} does not exist.'.format(
                                fname, i+1, N))
                            flist.remove(fname)

        # # load case matrix data to get descriptive case naming
        # if self.fname_case_matrix == '':
        #     print('Warning! No case matrix file provided, no case descriptions will be provided.')
        #     self.case_desc = ['Case ID %d' % i for i in range(M)]
        # else:
        #     cases = load_case_matrix(self.fname_case_matrix)
        #     self.case_desc = get_dlc_label(cases, include_seed=True)

        # get unique file namebase for datasets
        self.namebase = []
        if len(self.dataset_names) > 0:
            # use filename safe version of dataset names
            self.namebase = ["".join([c for c in name if c.isalpha() or c.isdigit() or c in [
                                     '_', '-']]).rstrip() for i, name in zip(range(N), self.dataset_names)]
        elif len(self.OpenFAST_outfile_list) > 0:
            # use out file naming
            if isinstance(self.OpenFAST_outfile_list[0], list):
                self.namebase = ['_'.join(os.path.split(flist[0])[1].split('_')[:-1])
                             for flist in self.OpenFAST_outfile_list]
            else:
                self.namebase = ['_'.join(os.path.split(flist)[1].split('_')[:-1])
                             for flist in self.OpenFAST_outfile_list]
        
        # check that names are unique
        if not len(self.namebase) == len(set(self.namebase)):
            self.namebase = []
        # as last resort, give generic name
        if not self.namebase:
            if isinstance(self.OpenFAST_outfile_list[0], str):
                # Just one dataset name for single dataset
                self.namebase = ['dataset1']
            else:
                self.namebase = ['dataset' + ('{}'.format(i)).zfill(len(str(N-1))) for i in range(N)]
        
        # Run design comparison if filenames list has multiple lists
        if (len(self.OpenFAST_outfile_list) > 1) and (isinstance(self.OpenFAST_outfile_list[0], list)): 
            # Load stats and load rankings for design comparisons
            stats, load_rankings = self.design_comparison(self.OpenFAST_outfile_list)
        
        else:
            # Initialize Analysis
            loads_analysis = Analysis.Loads_Analysis()
            loads_analysis.verbose = self.verbose
            loads_analysis.t0 = self.t0
            loads_analysis.tf = self.tf
            loads_analysis.DEL_info = self.DEL_info
            loads_analysis.ranking_stats = self.ranking_stats
            loads_analysis.ranking_vars = self.ranking_vars

            # run analysis in parallel
            if self.parallel_analysis:
                pool = mp.Pool(self.parallel_cores)
                try:
                    stats_separate = pool.map(
                        partial(loads_analysis.full_loads_analysis, get_load_ranking=False), self.OpenFAST_outfile_list)
                except:
                    stats_separate = pool.map(partial(loads_analysis.full_loads_analysis, get_load_ranking=False), self.OpenFAST_outfile_list[0])
                pool.close()
                pool.join()

                # Re-sort into the more "standard" dictionary/dataframe format we like
                stats = [pdTools.dict2df(ss).unstack() for ss in stats_separate]
                dft = pd.DataFrame(stats)
                dft = dft.reorder_levels([2, 0, 1], axis=1).sort_index(axis=1, level=0)
                stats = pdTools.df2dict(dft)

                # Get load rankings after stats are loaded
                load_rankings = loads_analysis.load_ranking(stats,
                                            names=self.dataset_names, get_df=False)
           
            # run analysis in serial
            else:
                # Initialize Analysis
                loads_analysis = Analysis.Loads_Analysis()
                loads_analysis.verbose = self.verbose
                loads_analysis.t0 = self.t0
                loads_analysis.tf = self.tf
                loads_analysis.ranking_stats = self.ranking_stats
                loads_analysis.ranking_vars = self.ranking_vars
                loads_analysis.DEL_info = self.DEL_info
                stats, load_rankings = loads_analysis.full_loads_analysis(self.OpenFAST_outfile_list, get_load_ranking=True)

        if self.save_SummaryStats:
            if isinstance(stats, dict):
                fname = self.namebase[0] + '_stats.yaml'
                if self.verbose:
                    print('Saving {}'.format(fname))
                save_yaml(self.results_dir, fname, stats)
            else:
                for namebase, st in zip(self.namebase, stats):
                    fname = namebase + '_stats.yaml'
                    if self.verbose:
                        print('Saving {}'.format(fname))
                    save_yaml(self.results_dir, fname, st)
        if self.save_LoadRanking:
            if isinstance(load_rankings, dict):
                fname = self.namebase[0] + '_LoadRanking.yaml'
                if self.verbose:
                    print('Saving {}'.format(fname))
                save_yaml(self.results_dir, fname, load_rankings)
            else:
                for namebase, lr in zip(self.namebase, load_rankings):
                    fname = namebase + '_LoadRanking.yaml'
                    if self.verbose:
                        print('Saving {}'.format(fname))
                    save_yaml(self.results_dir, fname, lr)


        return stats, load_rankings

    def design_comparison(self, filenames):
        '''
        Compare design runs

        Parameters:
        ----------
        filenames: list
            list of lists, where the inner lists are of equal length. 

        Returns:
        --------
        stats: dict
            dictionary of summary statistics data
        load_rankings: dict
            dictionary of load rankings
        '''


        # Make sure datasets are the same length
        ds_len = len(filenames[0])
        if any(len(dataset) != ds_len for dataset in filenames):
            raise ValueError('The datasets for filenames corresponding to the design comparison should all be the same size.')

        fnames = np.array(filenames).T.tolist()
        # Setup FAST_Analysis preferences
        loads_analysis = Analysis.Loads_Analysis()
        loads_analysis.verbose=self.verbose
        loads_analysis.t0 = self.t0
        loads_analysis.tf = self.tf
        loads_analysis.ranking_vars = self.ranking_vars
        loads_analysis.ranking_stats = self.ranking_stats
        loads_analysis.DEL_info = self.DEL_info
        
        if self.parallel_analysis: # run analysis in parallel
            # run analysis
            pool = mp.Pool(self.parallel_cores)
            stats_separate = pool.map(partial(loads_analysis.full_loads_analysis, get_load_ranking=False), fnames)
            pool.close()
            pool.join()
        
            # Re-sort into the more "standard" dictionary/dataframe format we like
            stats = [pdTools.dict2df(ss).unstack() for ss in stats_separate]
            dft = pd.DataFrame(stats)
            dft = dft.reorder_levels([2, 0, 1], axis=1).sort_index(axis=1, level=0)
            stats = pdTools.df2dict(dft)

            # Get load rankings after stats are loaded
            load_rankings = loads_analysis.load_ranking(stats) 

        else: # run analysis in serial
            stats = []
            load_rankings = []
            for file_sets in filenames:
                st, lr = loads_analysis.full_loads_analysis(file_sets, get_load_ranking=True)
                stats.append(st)
                load_rankings.append(lr)
            



        return stats, load_rankings

def get_windspeeds(case_matrix, return_df=False):
    '''
    Find windspeeds from case matrix

    Parameters:
    ----------
    case_matrix: dict
        case matrix data loaded from wisdem.aeroelasticse.Util.FileTools.load_yaml
    
    Returns:
    --------
    windspeed: list
        list of wind speeds
    seed: seed
        list of wind seeds
    IECtype: list
        list of IEC types 
    case_matrix: pd.DataFrame
        case matrix dataframe with appended wind info
    '''
    if isinstance(case_matrix, dict):
        cmatrix = case_matrix
    elif isinstance(case_matrix, pd.DataFrame):
        cmatrix = case_matrix.to_dict('list')
    else:
        raise TypeError('case_matrix must be a dict or pd.DataFrame.')


    windspeed = []
    seed = []
    IECtype = []
    # loop through and parse each inflow filename text entry to get wind and seed #
    for fname in  cmatrix[('InflowWind','Filename')]:
        if '.bts' in fname:
            obj = fname.split('U')[-1].split('_')
            obj2 = obj[1].split('Seed')[-1].split('.bts')
            windspeed.append(float(obj[0]))
            seed.append(float(obj2[0]))
            if 'NTM' in fname:
                IECtype.append('NTM')
            elif 'ETM' in fname:
                IECtype.append('NTM')
        elif 'ECD' in fname:
            obj = fname.split('U')[-1].split('.wnd')
            windspeed.append(float(obj[0]))
            seed.append([])
            IECtype.append('ECD')
        elif 'EWS' in fname:
            obj = fname.split('U')[-1].split('.wnd')
            windspeed.append(float(obj[0]))
            seed.append([])
            IECtype.append('EWS')
        
    if return_df:
        case_matrix = pd.DataFrame(case_matrix)
        case_matrix[('InflowWind','WindSpeed')] = windspeed
        case_matrix[('InflowWind','Seed')] = seed
        case_matrix[('InflowWind','IECtype')] = IECtype
        
        return windspeed, seed, IECtype, case_matrix
    
    else:
        return windspeed, seed, IECtype


def save_yaml(outdir, fname, data_out):
    ''' Save yaml file - ripped from WISDEM 
    
    Parameters:
    -----------
    outdir: str
        directory to save yaml
    fname: str
        filename for yaml
    data_out: dict
        data to dump to yaml
    '''
    fname = os.path.join(outdir, fname)

    if not os.path.isdir(outdir):
        os.makedirs(outdir)

    f = open(fname, "w")
    yaml = ry.YAML()
    yaml.default_flow_style = None
    yaml.width = float("inf")
    yaml.indent(mapping=4, sequence=6, offset=3)
    yaml.dump(data_out, f)


def load_yaml(fname_input, package=0):
    ''' Import a .yaml file - ripped from WISDEM

    Parameters:
    -----------
    fname_input: str
        yaml file to load
    package: bool
        0 = yaml, 1 = ruamel
    '''
    if package == 0:
        with open(fname_input) as f:
            data = yaml.safe_load(f)
        return data

    elif package == 1:
        with open(fname_input, 'r') as myfile:
            text_input = myfile.read()
        myfile.close()
        ryaml = ry.YAML()
        return dict(ryaml.load(text_input))



'''
Some tools to ease management of batch analysis data in pandas
'''
import pandas as pd
from pCrunch import Processing

def dict2df(sumstats, names=None):
    '''
    Build pandas datafrom from list of summary statistics

    Inputs:
    -------
    sumstats: list
        List of the dictionaries loaded from post_process.load_yaml
    names: list, optional
        List of names for each run. len(sumstats)=len(names)

    Returns:
    --------
    df: pd.DataFrame
        pandas dataframe 
    '''
    if isinstance(sumstats, list):
        if not names:
            names = ['dataset_' + str(i) for i in range(len(sumstats))]
        data_dict = {(name, outerKey, innerKey): values 
                        for name, sumdata in zip(names, sumstats)
                        for outerKey, innerDict in sumdata.items() 
                        for innerKey, values in innerDict.items()}
    else:
        data_dict = {(outerKey, innerKey): values 
                        for outerKey, innerDict in sumstats.items()
                        for innerKey, values in innerDict.items()}

    # Make dataframe
    df = pd.DataFrame(data_dict)

    return df


def df2dict(df):
    '''
    Build dictionary from pandas dataframe

    Parameters:
    -----------
    df: DataFrame
        DataFrame with summary stats (probably). Cannot have len(multiindex) > 3

    Returns:
    --------
    dfd: dict
        dictionary containing re-structured dataframe inputs
    '''
    if len(df.columns[0]) == 3:
        dfd = [{level: df.xs(dset, axis=1).xs(level, axis=1).to_dict('list')
              for level in df.columns.levels[1]} 
              for dset in df.columns.levels[0]]
    elif len(df.columns[0]) == 2:
        dfd = {level: df.xs(dset, axis=1).xs(level, axis=1).to_dict('list')
              for level in df.columns.levels[0]} 
    elif len(df.columns[0]) == 1:
        dfd = df.to_dict('list')
    else:
        raise TypeError('Converting DataFrames with multiindex > 3 to dictionaries is not supported')

    return dfd

def yaml2df(filename, names=[]):
    '''
    Read yaml containing summary stats into dataframe

    Parameters:
    -------
    filename:
        Name of yaml file to load. 

    '''

    data_dict = Processing.load_yaml('test.yaml', package=0)

    level = data_dict
    li = 0 # level index
    while isinstance(level, dict):
        li = li + 1
        if isinstance(level, dict):
            level = level[list(level.keys())[0]]

    data_list = []
    if li == 3:
        for key in data_dict.keys():
            data_list.append(data_dict[key])
        if not names:
            names = ['dataset_' + str(i) for i in range(len(data_list))]

    elif li == 2:
        data_list.append(data_dict)
        names = []
    else:
        raise TypeError('{} is improperly structured for yaml2df.'.format(filename))

    df = dict2df(data_list, names=names)

    return df



from __future__ import print_function
import os, sys, time, shutil
from functools import partial
import multiprocessing as mp
import numpy as np
import ruamel_yaml as ry
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import fatpack # 3rd party module used for rainflow counting

from scipy.interpolate import PchipInterpolator
from ROSCO_toolbox.utilities import FAST_IO

from pCrunch import pdTools

class Loads_Analysis(object):
    '''
    Contains analysis tools to post-process OpenFAST output data. Most methods are written to support
    single instances of output data to ease parallelization. 

    Methods:
    --------
    full_loads_analysis
    summary_stats
    load_ranking
    fatigue
    '''
    def __init__(self, **kwargs):

        # Analysis time range
        self.t0 = None
        self.tf = None
        # Desired channels for analysis
        self.channel_list = []
        self.channels_magnitude = {}

        self.channels_extreme_table = []

        # Load Ranking 
        self.ranking_vars = [["RotSpeed"],
                             ["TipDxc1", "TipDxc2", "TipDxc3"],
                             ["TipDyc1", "TipDyc2", "TipDyc3"],
                             ['RootMyb1', 'RootMyb2', 'RootMyb3'],
                             ['RootMxb1', 'RootMxb2', 'RootMxb3'],
                             ['TwrBsFyt']
                             ]  # List of lists
        self.ranking_stats = ['max',
                              'max',
                              'max',
                              'max',
                              'max',
                              ]  # should be same length as ranking_vars

        self.DEL_info = None #[('RootMyb1', 10), ('RootMyb2', 10), ('RootMyb3', 10)]
        # verbose?
        self.verbose=False

        for k, w in kwargs.items():
            try:
                setattr(self, k, w)
            except:
                pass

        super(Loads_Analysis, self).__init__()

    def full_loads_analysis(self, filenames, get_load_ranking=True, return_FastData=False):
        '''
        Load openfast data - get statistics - get load ranking - return data
        NOTE: Can be called to run in parallel if get_load_ranking=False (see Processing.batch_processing)

        Parameters:
        -------
        filenames: list
            List of filenames to load and analyse
        get_load_ranking: bool, optional
            Get the load ranking for all cases
        return_FastData, bool
            Return a dictionary or list constaining OpenFAST output data

        Outputs:
        --------
        sum_stats: dict
            dictionary of summary statistics
        load_rankings: dict
            dictionary of load rankings
        fast_data: list or dict
            list or dictionary containing OpenFAST output data
        '''
        # Load openfast data
        fast_io = FAST_IO()
        fast_data = fast_io.load_FAST_out(filenames, tmin=self.t0, tmax=self.tf, verbose=self.verbose)

        # Get summary statistics
        sum_stats = self.summary_stats(fast_data)

        # Get load rankings
        if get_load_ranking:
            load_rankings = self.load_ranking(sum_stats)


        if return_FastData:
            return sum_stats, fast_data
        if get_load_ranking: 
            return sum_stats, load_rankings
        if return_FastData and get_load_ranking:
            return sum_stats, load_rankings, fast_data
        else:
            return sum_stats

    def summary_stats(self, fast_data, channel_list=[]):
        '''
        Get summary statistics from openfast output data. 

        Parameters:
        ----------
        fast_data: list
            List of dictionaries containing openfast output data (returned from ROSCO_toolbox.FAST_IO.load_output)
        channel_list: list
            list of channels to collect data from. Defaults to all


        Returns:
        -------
        data_out: dict
            Dictionary containing summary statistics
        fast_outdata: dict, optional
            Dictionary of all OpenFAST output data. Only returned if return_data=true
        '''
        sum_stats     = {}
        extreme_table = {}

        for fd in fast_data:
            if self.verbose:
                print('Processing data for {}'.format(fd['meta']['name']))

            # Build channel list if it isn't input
            if channel_list == []:
                channel_list = list(fd.keys())

            # Process Data
            for channel in channel_list + list(self.channels_magnitude.keys()):
                if channel == 'meta':
                    if 'meta' not in sum_stats.keys():
                        sum_stats['meta'] = {}
                        sum_stats['meta']['name'] = []
                        sum_stats['meta']['filename'] = []
                    # save some meta data
                    sum_stats['meta']['name'] = fd['meta']['name']
                    sum_stats['meta']['filename'] = fd['meta']['filename']

                elif channel != 'Time' and channel in channel_list+list(self.channels_magnitude.keys()):
                    # try:

                    if channel in channel_list:
                        y_data = fd[channel]
                    elif channel in self.channels_magnitude.keys():
                        # calculate magnitude of a vector
                        vector = np.array([fd[var] for var in self.channels_magnitude[channel]])
                        n_dim, n_t = np.shape(vector)
                        y_data = np.array([np.sqrt(np.sum([vector[i,j]**2. for i in range(n_dim)])) for j in range(n_t)])

                    if channel not in sum_stats.keys():
                        sum_stats[channel] = {}
                        sum_stats[channel]['min'] = []
                        sum_stats[channel]['max'] = []
                        sum_stats[channel]['std'] = []
                        sum_stats[channel]['mean'] = []
                        sum_stats[channel]['abs'] = []
                        sum_stats[channel]['integrated'] = []

                    # calculate summary statistics
                    sum_stats[channel]['min'].append(float(min(y_data)))
                    sum_stats[channel]['max'].append(float(max(y_data)))
                    sum_stats[channel]['std'].append(float(np.std(y_data)))
                    sum_stats[channel]['mean'].append(float(np.mean(y_data)))
                    sum_stats[channel]['abs'].append(float(max(np.abs(y_data))))
                    sum_stats[channel]['integrated'].append(float(np.trapz(fd['Time'], y_data)))

                    if len(self.channels_extreme_table) > 0:
                        # outputting user specifed channels at the time where the maximum value occurs
                        if channel not in extreme_table:
                            extreme_table[channel] = []

                        extreme_table_i = {}
                        idx_max = np.argmax(y_data)
                        for var in self.channels_extreme_table:
                            extreme_table_i[var] = {}
                            extreme_table_i[var]['time'] = fd['Time'][idx_max]
                            extreme_table_i[var]['val']  = fd[var][idx_max]
                            
                        extreme_table[channel].append(extreme_table_i)

                    # except ValueError:
                    #     print('Error loading data from {}.'.format(channel))
                    # except:
                    #     print('{} is not in available OpenFAST output data.'.format(channel))

            # if self.channels_magnitude:
            #     for channel in self.channels_magnitude.keys():


            # Add DELS to summary stats
            if self.DEL_info:
                for channel, m in self.DEL_info:
                    if channel not in sum_stats.keys():
                        print('Cannot get DELs for {} because it does not exist in output data.'.format(channel))
                        break
                    if 'DEL' not in sum_stats[channel].keys():
                        sum_stats[channel]['DEL'] = []
                    
                    dfDEL = self.get_DEL([fd], [(channel, m)], t=fd['Time'][-1])
                    sum_stats[channel]['DEL'].append(float(dfDEL[channel][0]))


        if len(self.channels_extreme_table) > 0:
            return sum_stats, extreme_table
        else:
            return sum_stats


    def load_ranking(self, stats, names=[], get_df=False):
        '''
        Find load rankings for desired signals

        Parameters:
        -------
        stats: dict, list, pd.DataFrame
            summary statistic information
        ranking_stats: list
            desired statistics to rank for load ranking (e.g. ['max', 'std'])
        ranking_vars: list
            desired variables to for load ranking (e.g. ['GenTq', ['RootMyb1', 'RootMyb2', 'RootMyb3']]) 
        names: list of strings, optional
            names corresponding to each dataset
        get_df: bool, optional
            Return pd.DataFrame of data?
        
        Returns:
        -------
        load_ranking: dict
            dictionary containing load rankings
        load_ranking_df: pd.DataFrame
            pandas DataFrame containing load rankings
        '''
        
        # Make sure stats is in pandas df
        if isinstance(stats, dict):
            stats_df = pdTools.dict2df([stats], names=names)
        elif isinstance(stats, list):
            stats_df = pdTools.dict2df(stats, names=names)
        elif isinstance(stats, pd.DataFrame):
            stats_df = stats
        else:
            raise TypeError('Input stats is must be a dictionary, list, or pd.DataFrame containing OpenFAST output statistics.')


        # Ensure naming consitency
        if not names:
            names = list(stats_df.columns.levels[0])

        if self.verbose:
            print('Calculating load rankings.')
            
        # Column names to search in stats_df
        #  - [name, variable, stat],  i.e.['DLC1.1','TwrBsFxt','max']
        cnames = [pd.MultiIndex.from_product([names, var, [stat]])
                for var, stat in zip(self.ranking_vars, self.ranking_stats)]

        rank__ascending = False
        # Collect load rankings
        collected_rankings = []
        for col in cnames:
            # Set column names for dataframe
            mi_name = list(col.levels[0])
            mi_stat = col.levels[2]  # length = 1
            mi_idx = col.levels[2][0] + '_case_idx'
            if len(col.levels[1]) > 1:
                mi_var = [col.levels[1][0][:-1]]
            else:
                mi_var = list(col.levels[1])
            mi_colnames = pd.MultiIndex.from_product([mi_name, mi_var, [mi_idx, mi_stat[0]]])

            # Check for valid stats
            for c in col:
                if c not in list(stats_df.columns.values):
                    print('WARNING: {} does not exist in statistics.'.format(c))
                    col = col.drop(c)
                    # raise ValueError('{} does not exist in statistics'.format(c))
            # Go to next case if no [stat, var] exists in this set
            if len(col) == 0:
                continue
            # Extract desired variables from stats dataframe
            if mi_stat in ['max', 'abs']:
                var_df = stats_df[col].max(axis=1, level=0)
                rank__ascending = False
            elif mi_stat in ['min']:
                var_df = stats_df[col].min(axis=1, level=0)
                rank__ascending = True
            elif mi_stat in ['mean', 'std']:
                var_df = stats_df[col].mean(axis=1, level=0)
                rank__ascending = False

            # Combine ranking dataframes for each dataset
            var_df_list = [var_df[column].sort_values(
                ascending=rank__ascending).reset_index() for column in var_df.columns]
            single_lr = pd.concat(var_df_list, axis=1)
            single_lr.columns = mi_colnames
            collected_rankings.append(single_lr)

        # Combine dataframes for each case
        load_ranking_df = pd.concat(collected_rankings, axis=1).sort_index(axis=1)
        # Generate dict of info
        load_ranking = pdTools.df2dict(load_ranking_df)

        if get_df:
            return load_ranking, load_ranking_df
        else:
            return load_ranking

    def get_DEL(self, fast_data, chan_info, binNum=100, t=600):
        """ Calculates the short-term damage equivalent load of multiple variables
        
        Parameters: 
        -----------
        fast_data: list
            List of dictionaries containing openfast output data (returned from ROSCO_toolbox.FAST_IO.load_output)
        
        chan_info : list, tuple
            tuple/list containing channel names to be analyzed and corresponding fatigue slope factor "m"
            ie. ('TwrBsFxt',4)
        
        binNum : int
            number of bins for rainflow counting method (minimum=100)
        
        t : float/int
            Used to control DEL frequency. Default for 1Hz is 600 seconds for 10min data
        
        Outputs:
        -----------
        dfDEL : pd.DataFrame
            Damage equivalent load of each specified variable for one fast output file  
        """
        # check data types
        assert isinstance(fast_data, (list)), 'fast_data must be of type list'
        assert isinstance(chan_info, (list,tuple)), 'chan_info must be of type list or tuple'
        assert isinstance(binNum, (float,int)), 'binNum must be of type float or int'
        assert isinstance(t, (float,int)), 't must be of type float or int'

        # create dictionary from chan_dict
        dic = dict(chan_info)

        # pre-allocate list
        dflist = []
        names = []

        for fd in fast_data:
            dlist = [] # initiate blank list every loop
            # loop through channels and apply corresponding fatigue slope
            for var in dic.keys():
                if self.verbose:
                    print('Calculating DEL for {} in {}'.format(var, fd['meta']['name']))
                # find rainflow ranges
                ranges = fatpack.find_rainflow_ranges(fd[var])

                # find range count and bin
                Nrf, Srf = fatpack.find_range_count(ranges,binNum)

                # get DEL
                DELs = Srf**dic[var] * Nrf / t
                DEL = DELs.sum() ** (1/dic[var])
                dlist.append(DEL)
            # append DEL values for each channel to master list
            dflist.append(dlist)

            # create dataframe to return
            dfDEL = pd.DataFrame(np.transpose(dflist))
            dfDEL = dfDEL.T
            dfDEL.columns = dic.keys()
            
            #save simulation names
            names.append(fd['meta']['name'])
        dfDEL['Case_Name'] = names
        
        return dfDEL

class Power_Production(object):
    '''
    Class to generate power production stastics
    '''
    def __init__(self, **kwargs):
        # Turbine parameters
        self.turbine_class = 2

        for k, w in kwargs.items():
            try:
                setattr(self, k, w)
            except:
                pass

        super(Power_Production, self).__init__()

    def prob_WindDist(self, windspeed, disttype='pdf'):
        ''' 
        Generates the probability of a windspeed given the cumulative distribution or probability
        density function of a Weibull distribution per IEC 61400.

        NOTE: This uses the range of wind speeds simulated over, so if the simulated wind speed range
        is not indicative of operation range, using this cdf to calculate AEP is invalid
        
        Parameters:
        -----------
        windspeed: float or list-like
            wind speed(s) to calculate probability of 
        disttype: str, optional      
            type of probability, currently supports CDF or PDF
        Outputs:
        ----------
        p_bin: list
            list containing probabilities per wind speed bin 
        '''
        if self.turbine_class in [1, 'I']:
            Vavg = 50 * 0.2
        elif self.turbine_class in [2, 'II']:
            Vavg = 42.5 * 0.2
        elif self.turbine_class in [3, 'III']:
            Vavg = 37.5 * 0.2

        # Define parameters
        k = 2 # Weibull shape parameter
        c = (2 * Vavg)/np.sqrt(np.pi) # Weibull scale parameter 

        if disttype.lower() == 'cdf':
            # Calculate probability of wind speed based on WeibulCDF
            wind_prob = 1 - np.exp(-(windspeed/c)**k)
        elif disttype.lower() == 'pdf':
            # Calculate probability of wind speed based on WeibulPDF
            wind_prob = (k/c) * (windspeed/c)**(k-1) * np.exp(-(windspeed/c)**k)
        else:
            raise ValueError('The {} probability distribution type is invalid'.format(disttype))
        
        return wind_prob

    def AEP(self, stats, windspeeds, U_pwr_curve=[], pwr_curve_vars=[]):
        '''
        Get AEPs for simulation cases

        TODO: Print/Save this someplace besides the console
    
        Parameters:
        ----------
        stats: dict, list, pd.DataFrame
            Dict (single case), list(multiple cases), df(single or multiple cases) containing
            summary statistics. 
        windspeeds: list-like
            List of wind speed values corresponding to each power output in the stats input 
            for a single dataset

        n_pwr_curve: list-like
            List of wind speed values to output power variables

        pwr_curve_vars: list of strings
            List of OpenFAST output channels to return the mean value as a function of wind speeds

        Returns:
        --------
        AEP: List
            Annual energy production corresponding to 
        '''
  
        # Make sure stats is in pandas df
        if isinstance(stats, dict):
            stats_df = pdTools.dict2df(stats)
        elif isinstance(stats, list):
            stats_df = pdTools.dict2df(stats)
        elif isinstance(stats, pd.DataFrame):
            stats_df = stats
        else: 
            raise TypeError('Input stats is must be a dictionary, list, or pd.DataFrame containing OpenFAST output statistics.')

        # Check windspeed length
        if len(windspeeds) == len(stats_df):
            ws = windspeeds
        elif int(len(windspeeds)/len(stats_df.columns.levels[0])) == len(stats_df):
            ws = windspeeds[0:len(stats_df)]
            print('WARNING: Assuming the input windspeed array is duplicated for each dataset.')
        else:
            raise ValueError(
                'Length of windspeeds is not the correct length for the input statistics.')

        # load power array
        if 'GenPwr' in stats_df.columns.levels[0]:
            pwr_array = stats_df.loc[:, ('GenPwr', 'mean')]
            pwr_array = pwr_array.to_frame()
        elif 'GenPwr' in stats_df.columns.levels[1]:
            pwr_array = stats_df.loc[:, (slice(None), 'GenPwr', 'mean')]
        else:
            raise ValueError("('GenPwr','Mean') does not exist in the input statistics.")
        
        # group and average powers by wind speeds
        pwr_array['windspeeds'] = ws 
        pwr_array = pwr_array.groupby('windspeeds').mean() 
        # find set of wind speeds
        ws_set = list(set(ws))
        # wind probability
        wind_prob = self.prob_WindDist(ws_set, disttype='pdf')
        # Calculate AEP
        AEP = np.trapz(pwr_array.T *  wind_prob, ws_set) * 8760

        # return power curves
        if len(pwr_curve_vars) > 0:
            performance_curves = {}
            if len(U_pwr_curve) > 0:
                performance_curves['U'] = U_pwr_curve
            else:
                performance_curves['U'] = ws_set

            for var in pwr_curve_vars:
                # get data
                if var in stats_df.columns.levels[0]:
                    perf_array = stats_df.loc[:, (var, 'mean')]
                    perf_array = perf_array.to_frame()
                elif var in stats_df.columns.levels[1]:
                    perf_array = stats_df.loc[:, (slice(None), var, 'mean')]
                else:
                    raise ValueError("(%s,'Mean') does not exist in the input statistics."%var)
                # average by wind speed
                perf_array['windspeeds'] = ws 
                perf_array = perf_array.groupby('windspeeds').mean()

                if len(U_pwr_curve) > 0:
                    spline = PchipInterpolator(ws_set, perf_array[var])
                    performance_curves[var] = spline(performance_curves['U']).flatten()
                else:
                    performance_curves[var] = perf_array[var]

        

        if len(pwr_curve_vars) > 0:
            return AEP, performance_curves
        else:
            return AEP

class wsPlotting(object):
    '''
    General plotting scripts.
    '''

    def __init__(self):
        pass

    def stat_curve(self, windspeeds, stats, plotvar, plottype, stat_idx=0, names=[]):
        '''
        Plot the turbulent power curve for a set of data. 
        Can be plotted as bar (good for comparing multiple cases) or line 

        Parameters:
        -------
        windspeeds: list-like
            List of wind speeds to plot
        stats: list, dict, or pd.DataFrame
            Dict (single case), list(multiple cases), df(single or multiple cases) containing
            summary statistics. 
        plotvar: str
            Type of variable to plot
        plottype: str
            bar or line 
        stat_idx: int, optional
            Index of datasets in stats to plot from
        
        Returns:
        --------
        fig: figure handle
        ax: axes handle
        '''

        # Check for valid inputs
        if isinstance(stats, dict):
            stats_df = pdTools.dict2df(stats)
            if any((stat_inds > 0) or (isinstance(stat_inds, list))):
                print('WARNING: stat_ind = {} is invalid for a single stats dictionary. Defaulting to stat_inds=0.')
                stat_inds = 0
        elif isinstance(stats, list):
            stats_df = pdTools.dict2df(stats)
        elif isinstance(stats, pd.DataFrame):
            stats_df = stats
        else:
            raise TypeError(
                'Input stats must be a dictionary, list, or pd.DataFrame containing OpenFAST output statistics.')

       
        # Check windspeed length
        if len(windspeeds) == len(stats_df):
            ws = windspeeds
        elif int(len(windspeeds)/len(stats_df.columns.levels[0])) == len(stats_df):
            ws = windspeeds[0:len(stats_df)]
        else:
            raise ValueError('Length of windspeeds is not the correct length for the input statistics')

        # Get statistical data for desired plot variable
        if plotvar in stats_df.columns.levels[0]:
            sdf = stats_df.loc[:, (plotvar, slice(None))].droplevel([0], axis=1)
        elif plotvar in stats_df.columns.levels[1]:
            sdf = stats_df.loc[:, (slice(None), plotvar, slice(None))].droplevel([1], axis=1)
        else:
            raise ValueError("{} does not exist in the input statistics.".format(plotvar))
        
        # Add windspeeds to data
        sdf['WindSpeeds']= ws
        # Group by windspeed and average each statistic (for multiple seeds)
        sdf = sdf.groupby('WindSpeeds').mean() 
        # Final wind speed values
        pl_windspeeds=sdf.index.values

        if plottype == 'bar':
            # Define mean and std dataframes
            means = sdf.loc[:, (slice(None), 'mean')].droplevel(1, axis=1)
            std = sdf.loc[:, (slice(None), 'std')].droplevel(1, axis=1)
            # Plot bar charts
            fig, ax = plt.subplots(constrained_layout=True)
            means.plot.bar(yerr=std, ax=ax, title=plotvar, capsize=2)
            ax.legend(names,loc='upper left')

        if plottype == 'line':
            # Define mean, min, max, and std dataframes
            means = sdf.loc[:, (sdf.columns.levels[0][stat_idx], 'mean')]
            smax = sdf.loc[:, (sdf.columns.levels[0][stat_idx], 'max')]
            smin = sdf.loc[:, (sdf.columns.levels[0][stat_idx], 'min')]
            std = sdf.loc[:, (sdf.columns.levels[0][stat_idx], 'std')]

            fig, ax = plt.subplots(constrained_layout=True)
            ax.errorbar(pl_windspeeds, means, [means - smin, smax - means],
                         fmt='k', ecolor='gray', lw=1, capsize=2)
            means.plot(yerr=std, ax=ax, 
                        capsize=2, lw=3, 
                        elinewidth=2, 
                        title=names[0] + ' - ' + plotvar)
            plt.grid(lw=0.5, linestyle='--')

        return fig, ax


    def distribution(self, fast_data, channels, caseid, names=None, kde=True):
        '''
        Distributions of data from desired fast runs and channels

        Parameters
        ----------
        fast_data: dict, list
            List or Dictionary containing OpenFAST output data from desired cases to compare
        channels: list
            List of strings of OpenFAST output channels e.g. ['RotSpeed','GenTq']
        caseid: list
            List of caseid's to compare
        names: list, optional
            Names of the runs to compare
        fignum: ind, (optional)
            Specified figure number. Useful to plot over previously made plot
        
        Returns:
        --------
        fig: figure handle
        ax: axes handle
        '''
        # Make sure input types allign
        if isinstance(fast_data, dict):
            fd = [fast_data]
        elif isinstance(fast_data, list):
            if len(caseid) == 1:
                fd = [fast_data[caseid[0]]]
            else:
                fd = [fast_data[case] for case in caseid]
        else:
            raise ValueError('fast_data is an improper data type')
            

        # if not names:
        #     names = [[]]*len(fd)

        for channel in channels:
            fig, ax = plt.subplots()
            for idx, data in enumerate(fd):
                # sns.kdeplot(data[channel], shade=True, label='case '+ str(idx))
                sns.distplot(data[channel], kde=kde, label='case ' + str(idx))  # For a histogram
                ax.set_title(channel + ' distribution')

                units = data['meta']['attribute_units'][data['meta']['channels'].index(channel)]
                ax.set_xlabel('{} [{}]'.format(channel, units))
                ax.grid(True)
            if names:
                ax.legend(names)
                
        return fig, ax

    def plot_load_ranking(self, load_rankings, case_matrix, classifier_type, 
                        classifier_names=[], n_rankings=10, caseidx_labels=False):
        '''
        Plot load rankings
        
        Parameters:
        -----------
        load_rankings: list, dict, or pd.DataFrame
            Dict (single case), list(multiple cases), df(single or multiple cases) containing
            load rankings. 
        case_matrix: dict or pdDataFrame
            Information mapping classifiers to load_rankings. 
            NOTE: the case matrix must have wind speeds in ('InflowWind','WindSpeeds') if you 
            wish to plot w.r.t. wind speed
        classifier_type: tuple or str
            classifier to denote load ranking cases. e.g. classifier_type=('IEC','DLC') will separate
            the load rankings by DLC type, assuming the case matrix is properly set up to map each
            DLC to the load ranking case
        classifier_names: list, optional
            Naming conventions for each classifier type for plotting purposes
        n_rankings: int, optional
            number of load rankings to plot
        caseidx_labels: bool, optional
            label x-axis with case index if True. If false, will try to plot with wind speed labels
            if they exist, then fall abck to case indeces.  

        TODO: Save figs
        '''

        # flag_DLC_name = False
        # n_rankings = 10
        # fig_ext = '.pdf'
        # font_size = 10
        # classifier_type = ('ServoDyn', 'DLL_FileName')
        # classifiers = list(set(cmw[classifier_type]))
        # classifier_names = ['ROSCO', 'legacy']
       
       # Check for valid inputs
        if isinstance(load_rankings, dict):
            load_ranking_df = pdTools.dict2df(load_rankings)
        elif isinstance(load_rankings, list):
            load_ranking_df = pdTools.dict2df(load_rankings)
        elif isinstance(load_rankings, pd.DataFrame):
            load_ranking_df = load_rankings
        else:
            raise TypeError(
                'Input stats must be a dictionary, list, or pd.DataFrame containing OpenFAST output statistics.')

        # Check multiindex size
        if len(load_ranking_df) == 2:
            load_ranking_df = pd.concat([load_ranking_df], keys=[dataset_0])

        # Check for classifier_names
        classifiers = list(set(case_matrix[classifier_type]))
        if not classifier_names:
            classifier_names = ['datatset_{}'.format(idx) for idx in range(len(classifiers))]

        # Check for wind speeds in case_matrix
        if not caseidx_labels:
            try:
                windspeeds = case_matrix[('InflowWind','WindSpeed')]
            except: 
                print('Unable to find wind speeds in case_matrix, plotting w.r.t case index')    
                caseidx_labels=True


        # Define a color map
        clrs = np.array([[127, 60, 141],
                        [17, 165, 121],
                        [57, 105, 172],
                        [242, 183, 1],
                        [231, 63, 116],
                        [128, 186, 90],
                        [230, 131, 16],
                        [256, 256, 256]]) / 256.

        # Get channel names
        channels = load_ranking_df.columns.levels[1]

        # initialize some variables
        colors = np.zeros((n_rankings, 3))
        labels = [''] * n_rankings
        labels_index = [''] * n_rankings
        fig_list = []
        ax_list = []
        # --- Generate plots ---
        for cidx, channel in enumerate(channels):
            # Pull out specific channel
            cdf = load_ranking_df.loc[:, (slice(None), channel, slice(None))].droplevel(1, axis=1)
            # put the load ranking from each dataset in a list so we can combine them
            cdf_list = [cdf[dataset] for dataset in cdf.columns.levels[0]] 
            chan_df = pd.concat(cdf_list) # combine all load rankings
            chan_stats = chan_df.columns.values # pull out the names of the columns
            chan_df.sort_values(by=chan_stats[0], ascending=False, inplace=True) # sort
            chan_df.reset_index(inplace=True, drop=True) # re-index

            # find colors and labels for plots
            for i in range(n_rankings):
                classifier = case_matrix[classifier_type][chan_df[chan_stats[1]][i]]
                colors[i, :] = clrs[min(len(clrs), classifiers.index(classifier))]

                if not caseidx_labels:
                    ws = windspeeds[chan_df[chan_stats[1]][i]]
                    labels[i] = classifier_names[classifiers.index(classifier)] + ' - ' + str(ws) + ' m/s'
                else:
                    labels[i] = classifier_names[classifiers.index(classifier)] + ' - Case ' + str(chan_df[chan_stats[1]][i])
        #         labels_index = ['case {}'.format(case) for case in chan_df[chan_stats[1]][0:n_rankings]]

            # make plot
            fig, ax = plt.subplots()
            chan_df[chan_stats[0]][0:n_rankings].plot.bar(color=colors)
            ax.set_ylabel(channel)
            ax.set_xticklabels(labels, rotation=45, ha='right')
            plt.draw()

            fig_list.append(fig)
            ax_list.append(ax)

        #     if case_idx_labels:
        #         ax.set_xlabel('DLC [-]', fontsize=font_size+2, fontweight='bold')
        # #         ax.set_xticklabels(np.arange(n_rankings), labels=labels)
        #         ax.set_xticklabels(labels)
        #     else:
        #         #         ax.set_xticklabels(np.arange(n_rankings), labels=labels)

            


        return fig_list, ax_list
